/*  PMG, a VOS clone for linux.
    Copyright (C) 2006 rainycat <rainy6144@gmail.com>.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA. */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>
#include <asoundlib.h>
#include <popt.h>
#include <openssl/evp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#define IDX_NONE ((unsigned) -1)

#define NKEY 7

/* A note to be played by the user */
typedef struct usernote_s {
  double rt_start; /* Time of the Note-on event in seconds from the start of the song */
  double tt_start; /* Time of the Note-on event in MIDI quarter-notes from the start of the song */
  double rt_stop, tt_stop; /* For long notes, it is {rt,tt}_end; for short notes, it is {rt,tt}_start.
			      Used when displaying notes and scoring. */
  double rt_end, tt_end; /* Time of the corresponding Note-off event.  Used when playing short notes. */
  unsigned char cmd, note_num, vol; /* Corresponding MIDI command */
  unsigned key, color; /* Key is 0-6 for do,re,mi,fa,so,la,ti */
  gboolean is_long;
  unsigned prev_idx, next_idx; /* The array index of the previous/next note on the same key;
				  IDX_NONE if there is none */
  unsigned prev_idx_long, next_idx_long;
  unsigned count, count_long; /* The number of (long) notes on this key, including this one */
} UserNote;

typedef struct inputevent_s {
  double rt;
  gboolean is_keypress;
  unsigned key;
} InputEvent;

typedef struct tempochange_s {
  double rt0, tt0; /* rt and tt at the time when the tempo change occurs */
  double qn_time; /* The duration of a MIDI quarter-note after that */
} TempoChange;

typedef struct midievent_s {
  double tt;
  unsigned char cmd, a, b; /* cmd is the status byte */
} MidiEvent;

/* Stores the MIDI initialization events in an MIDI track */
typedef struct miditrack_s {
  unsigned nevent;
  MidiEvent *events;
} MidiTrack;

/* Both BGM notes and notes played by the user */
typedef struct note_s {
  double tt_start, tt_end;
  unsigned cmd, note_num, vol;
  gboolean is_user;
} Note;

typedef struct notearray_s {
  unsigned num_note;
  Note *notes;
} NoteArray;

/* An iterator in the TempoChange array, used to accelerate the converstion between rt's and tt's. */
typedef struct tempoit_s {
  unsigned idx; /* The index of the current TempoChange being used */
} TempoIt;

/* A scheduled event */
typedef enum schedeventtype_e {
  SchedEventMidi, SchedEventNoteOn, SchedEventNoteOff, SchedEventUserNoteOff, SchedEventDemo,
  SchedEventRefreshScore, SchedEventAutoOn, SchedEventAutoOff
} SchedEventType;
typedef struct schedevent_s {
  double rt;
  SchedEventType type;
  unsigned src, idx; /* src is the track number or the note array number;
			idx is the index within the track or note array */
} SchedEvent;

typedef struct vossong_s {
  unsigned ntempo;
  TempoChange *tempos; /* The first TempoChange is the default one with qn_time=0.5 (120BPM) */
  double midi_unit_tt; /* The number of quarter-notes in a MIDI delta-time unit */
  unsigned ntrack;
  MidiTrack *tracks;
  unsigned num_note_array; /* Excluding the last one containing user notes */
  NoteArray *note_arrays;
  unsigned num_user_note;
  UserNote *user_notes;
  double max_rt;
  unsigned first_un_idxs[NKEY]; /* The first user note on each key */
  unsigned first_un_idxs_long[NKEY]; /* The first long user note */
  unsigned char hash[SHA_DIGEST_LENGTH];
} VosSong;

typedef struct comparerentry_s {
  unsigned last_idx; /* The index of the last user note used for matching */
  double score; /* Total score so far */
} ComparerEntry;

typedef struct comparer_s {
  double score0; /* Score when none of the user notes within range are used for matching */
  unsigned count0;
  unsigned nent;
  ComparerEntry *ents; /* Should be ordered so that both last_idx and score are strictly increasing */
} Comparer;

/* A marker is a circle showing the accuracy after a keypress */
typedef struct marker_s {
  double rt;
  double dt; /* dt = un->rt_start - cur_rt; dt > 0: too early */
} Marker;

typedef struct keystate_s {
  gboolean key_down; /* Whether the key on the keyboard is down or not */
  double key_down_rt;
  unsigned un_idx; /* For the actual note to play.
		      Only meaningful if key_down is true; IDX_NONE if the key_down had been ignored */
  Comparer cmp_press, cmp_release; /* One comparer for keypress events, another for key release events */
  double long_score, total_long_score;
  double long_rt; /* The score before long_rt has been counted into long_score and total_long_score. */
  GQueue *markers;
} KeyState;

typedef enum vosgamemode_e { VosGameModeNormal, VosGameModeAuto, VosGameModePlayDemo } VosGameMode;

typedef struct gamescore_s {
  double score, total_score;
} GameScore;

typedef struct vosgame_s {
  VosGameMode mode;
  GtkWidget *main_area;
  GtkLabel *state_label, *time_label, *score_label, *rate_label;
  VosSong *song;
  const char *user_name;
  snd_seq_t *seq;
  int seq_client_id, seq_port_id;
  int dest_client_id, dest_port_id;
  guint refresh_timeout_id, event_timeout_id;
  guint conn_key_press, conn_key_release, conn_expose, conn_delete;
  gboolean is_paused;
  double start_rt; /* The rt at start_time */
  GTimeVal start_time, pause_time;
  GArray *input_events_arr; /* Used when recording demos */
  InputEvent *in_evs; /* Used when playing demos */
  unsigned in_nev;
  double last_status_upd_rt;
  GameScore *recent_scores;
  double speed; /* vy per unit rt */
  GTree *sched; /* event scheduler */
  double pev_time_unit; /* time resolution of PEV files */
  GdkColor *colors;
  TempoIt it;
  unsigned cur_un_idxs[NKEY]; /* Used for speeding up the search in the user notes array;
				 Always points to a note on that key, unless there is no note on this key,
				 in which case the value is IDX_NONE */
  unsigned cur_un_idxs_long[NKEY];
  KeyState keys[NKEY];
} VosGame;

/* Information in the header of input events */
#define PEV_MAGIC 0xb10cde81U
/* Header types */
#define PEV_HEADER_END 0
#define PEV_HASH_SHA1 1
#define PEV_TEMPO_FACTOR 2 /* Skipped if this is 1 */
#define PEV_SPEED 3 /* NOTE: this might be bogus if speed can be changed in-game */
#define PEV_TIME_UNIT 4
#define PEV_DATE 5
#define PEV_USER_NAME 6
/* Event types */
#define PEV_EV_END 0
#define PEV_EV_PRESS 1
#define PEV_EV_RELEASE 8 /* The next event type should be 15 */

static double score_dt0 = 0.5; /* The amount of error, above which the lowest score is given */
static int fast_mode = FALSE; /* In fast mode, we mainly just calculate the scores */
static double default_speed = 1.0;
static double tempo_factor = 1.0;
static double default_pev_time_unit = 0.001;
static const char *pev_in_fname = NULL;
static const char *pev_out_fname = NULL;
static const char *midi_port_spec = "128:0";
static const double score_refresh_period = 0.25;
static const unsigned recent_scores_nperiod = 8;

static const struct poptOption popt_table[] = {
  { "port", 'p', POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT, &midi_port_spec, 0,
    "Set the MIDI sequencer port", "CLIENT:PORT" },
  { "speed", 's', POPT_ARG_DOUBLE | POPT_ARGFLAG_SHOW_DEFAULT, &default_speed, 0,
    "Set the speed of the notes falling down", "SPEED" },
  { "dt0", 'd', POPT_ARG_DOUBLE | POPT_ARGFLAG_SHOW_DEFAULT, &score_dt0, 0,
    "Set the maximum allowed error in seconds", "TIME" },
  { "tempo", 'T', POPT_ARG_DOUBLE, &tempo_factor, 0, "Multiply the tempo by FACTOR", "FACTOR" },
  { "record", 'o', POPT_ARG_STRING, &pev_out_fname, 0, "Record into demo file FNAME.pev", "FNAME.pev" },
  { "playdemo", 'i', POPT_ARG_STRING, &pev_in_fname, 'i', "Play demo file FNAME.pev", "FNAME.pev" },
  { "fast", 'f', POPT_ARG_NONE, &fast_mode, 0, "Fast mode", NULL },
  { "auto", 'A', POPT_ARG_NONE, NULL, 'A', "Automatic mode", NULL },
  POPT_AUTOHELP
  POPT_TABLEEND };

#define GET_WIDGET(gx, obj, name, type)			\
do {							\
  (obj) = type(glade_xml_get_widget((gx), (name)));	\
  g_assert((obj) != NULL);				\
 } while (0)

#define COLOR_BLACK 0
#define COLOR_WHITE 1
#define COLOR_KEYPRESS_START 2
#define COLOR_KEYPRESS 10
#define COLOR_NOTE_START 11
#define COLOR_NOTE_END 42
#define COLOR_MARKER_START 43
#define COLOR_MARKER_BEST 43
#define COLOR_MARKER_WORST 59
#define COLOR_MARKER_END 74

#define NCOLOR 75

static const char *glade_fname = "pmg.glade";
static const double vx_min = -0.1, vx_max = 7.1, vx_mid = 3.5, vy_min = -1.0, vy_max = 0.5;
static const unsigned note_thickness = 4; /* In pixels */

/* Whether the length of short notes are automatically lengthened to the actual note length */
static const gboolean short_note_auto_len = TRUE;

/* Returns t-t0 in seconds */
static double timeval_diff(GTimeVal *t0, GTimeVal *t)
{
  return (t->tv_sec - t0->tv_sec) + 1e-6 * (t->tv_usec - t0->tv_usec);
}

#define BUF_SIZE 4096
static void hash_file(FILE *file, unsigned char *hash, unsigned hash_len)
{
  static char buf[BUF_SIZE];
  unsigned cur_len;
  EVP_MD_CTX mdctx;
  const EVP_MD *md = EVP_sha1();
  int result;

  g_assert(EVP_MD_size(md) == (int) hash_len);
  result = EVP_DigestInit(&mdctx, md); g_assert(result == 1);
  result = fseek(file, 0, SEEK_SET); g_assert(result == 0);
  while (1) {
    result = fread(buf, 1, BUF_SIZE, file); g_assert(result >= 0); cur_len = result;
    if (cur_len == 0) break;
    result = EVP_DigestUpdate(&mdctx, buf, cur_len); g_assert(result == 1);
  }
  result = EVP_DigestFinal(&mdctx, hash, NULL); g_assert(result == 1);
}
#undef BUF_SIZE

static void tempoit_init(VosSong *song, TempoIt *it)
{
  it->idx = 0;
}

static double tempoit_tt_to_rt(VosSong *song, TempoIt *it, double tt)
{
  const TempoChange *tc;
  
  while (it->idx > 0 && tt < song->tempos[it->idx].tt0) it->idx--;
  while (it->idx + 1 < song->ntempo && tt >= song->tempos[it->idx+1].tt0) it->idx++;
  tc = &song->tempos[it->idx];
  return tc->rt0 + (tt - tc->tt0) * tc->qn_time;
}

static double tempoit_rt_to_tt(VosSong *song, TempoIt *it, double rt)
{
  const TempoChange *tc;

  while (it->idx > 0 && rt < song->tempos[it->idx].rt0) it->idx--;
  while (it->idx + 1 < song->ntempo && rt >= song->tempos[it->idx+1].rt0) it->idx++;
  tc = &song->tempos[it->idx];
  return tc->tt0 + (rt - tc->rt0) / tc->qn_time;
}

static gint event_compare(gconstpointer pa, gconstpointer pb, gpointer user_data)
{
  const SchedEvent *a = (SchedEvent *) pa, *b = (SchedEvent *) pb;
  if (a->rt != b->rt) return (a->rt > b->rt) - (a->rt < b->rt);
  else return (a > b) - (a < b);
}

static GTree *new_event_scheduler(void)
{
  return g_tree_new_full(event_compare, NULL, NULL, g_free);
}

static void free_event_scheduler(GTree *sched)
{
  g_tree_destroy(sched);
}

static void sched_event(GTree *sched, double rt, SchedEventType type, unsigned src, unsigned idx)
{
  SchedEvent *ev = g_new0(SchedEvent, 1);
  ev->rt = rt; ev->type = type; ev->src = src; ev->idx = idx;
  g_tree_insert(sched, ev, ev);
}

static gboolean get_first_event_each(gpointer key, gpointer value, gpointer data)
{
  SchedEvent **pev = (SchedEvent **) data;
  *pev = (SchedEvent *) key;
  return TRUE; /* stop traversal */
}

static SchedEvent *get_first_event(GTree *sched)
{
  SchedEvent *ev = NULL;
  g_tree_foreach(sched, get_first_event_each, &ev);
  return ev;
}

static void remove_event(GTree *sched, SchedEvent *ev)
{
  g_tree_remove(sched, ev);
}

static void free_vos_song(VosSong *song)
{
  unsigned i;
  
  g_free(song->tempos);
  for (i = 0; i < song->ntrack; i++) g_free(song->tracks[i].events);
  g_free(song->tracks);
  for (i = 0; i < song->num_note_array; i++) g_free(song->note_arrays[i].notes);
  g_free(song->note_arrays);
  g_free(song->user_notes);
  g_free(song);
}

static unsigned midi_read_u8(FILE *file)
{
  int result = getc(file);
  g_assert(result != EOF);
  return (unsigned) result;
}

static unsigned midi_read_u16(FILE *file)
{
  guint16 tmp;
  int result;

  result = fread(&tmp, 2, 1, file); g_assert(result == 1);
  return GUINT16_FROM_BE(tmp);
}

static unsigned midi_read_u32(FILE *file)
{
  guint32 tmp;
  int result;

  result = fread(&tmp, 4, 1, file); g_assert(result == 1);
  return GUINT32_FROM_BE(tmp);
}

static unsigned midi_read_vl(FILE *file)
{
  int result;
  unsigned x = 0;

  while (1) {
    result = getc(file); g_assert(result != EOF);
    x |= (result & 0x7f);
    if (result & 0x80) x <<= 7;
    else break;
  }
  return x;
}

static void vosfile_read_midi(VosSong *song, FILE *file, unsigned mid_ofs, unsigned mid_len)
{
  int result;
  char magic[4];
  unsigned header_len, fmt, ppqn, i;
  GArray *tempo_arr;
  TempoChange last_tc;
  
  result = fseek(file, mid_ofs, SEEK_SET); g_assert(result == 0);
  result = fread(magic, 4, 1, file); g_assert(result == 1);
  g_assert(memcmp(magic, "MThd", 4) == 0);
  header_len = midi_read_u32(file); g_assert(header_len == 6);
  fmt = midi_read_u16(file); g_assert(fmt == 1);
  song->ntrack = midi_read_u16(file); g_assert(song->ntrack > 0);
  ppqn = midi_read_u16(file); g_assert((ppqn & 0x8000) == 0); /* Actually a ppqn */
  song->midi_unit_tt = 1.0 / ppqn;

  song->tracks = g_new0(MidiTrack, song->ntrack);
  tempo_arr = g_array_new(FALSE, FALSE, sizeof(TempoChange));
  last_tc.rt0 = 0.0; last_tc.tt0 = 0.0; last_tc.qn_time = 0.5; /* 120 BPM, the default */
  g_array_append_val(tempo_arr, last_tc);
  for (i = 0; i < song->ntrack; i++) {
    MidiTrack *track = &song->tracks[i];
    GArray *event_arr = g_array_new(FALSE, FALSE, sizeof(MidiEvent));
    unsigned track_ofs, track_len;
    unsigned char cmd = 0; /* The status byte */
    unsigned char ch;
    unsigned time = 0; /* In MIDI units */
    gboolean is_eot = FALSE; /* Found "End of Track"? */

    result = fread(magic, 4, 1, file); g_assert(result == 1);
    g_assert(memcmp(magic, "MTrk", 4) == 0);
    track_len = midi_read_u32(file);
    track_ofs = ftell(file);
    while ((unsigned) ftell(file) < track_ofs + track_len) {
      MidiEvent ev;
      double tt;
      
      g_assert(! is_eot);
      memset(&ev, 0, sizeof(ev));
      time += midi_read_vl(file); tt = time * song->midi_unit_tt;
      ch = midi_read_u8(file);
      if (ch == 0xff) { /* Meta event */
	unsigned meta_type = midi_read_u8(file);
	unsigned meta_len = midi_read_vl(file);
	switch (meta_type) {
	default:
	  g_warning("Unknown meta event type 0x%02x", meta_type);
	  /* fall through */
	case 0x03: /* Sequence/Track name */
	case 0x21: /* MIDI port number */
	case 0x58: /* Time signature */
	case 0x59: /* Key signature */
	  result = fseek(file, meta_len, SEEK_CUR); g_assert(result == 0); /* skip over the data */
	  break;
	case 0x2f: /* End of track */
	  g_assert(meta_len == 0); is_eot = TRUE; break;
	case 0x51: /* Tempo */
	  {
	    unsigned char buf[3];
	    unsigned val;
	    TempoChange tc;
	    
	    g_assert(meta_len == 3); g_assert(tt >= last_tc.tt0);
	    result = fread(buf, meta_len, 1, file); g_assert(result == 1);
	    val = ((unsigned) buf[0] << 16) | ((unsigned) buf[1] << 8) | buf[2];
	    tc.qn_time = val * 1e-6 / tempo_factor;
	    tc.tt0 = tt; tc.rt0 = last_tc.rt0 + last_tc.qn_time * (tt - last_tc.tt0);
	    g_array_append_val(tempo_arr, tc);
	    /*	    g_message("Tempo change: time=%u tt0=%0.3f rt0=%0.3f", time, tc.tt0, tc.rt0); */
	    last_tc = tc;
	  }
	  break;
	}
	continue;
      } else if (ch == 0xf0) { /* SysEx event */
	unsigned sysex_len = midi_read_vl(file);
	/* Just ignore it */
	result = fseek(file, sysex_len, SEEK_CUR); g_assert(result == 0);
      } else {
	unsigned cmd_type;
	if (ch & 0x80) { cmd = ch; ch = midi_read_u8(file); }
	cmd_type = (cmd & 0x7f) >> 4; g_assert(cmd_type != 7);
	ev.cmd = cmd; ev.a = ch;
	/* Program Change and Channel Pressure messages have one status bytes, all others have two */
	if (! (cmd_type == 4 || cmd_type == 5)) ev.b = midi_read_u8(file);
      }
      ev.tt = tt;
      g_array_append_val(event_arr, ev);
    }
    g_assert(is_eot);
    g_assert((unsigned) ftell(file) == track_ofs + track_len);
    track->nevent = event_arr->len;
    track->events = (MidiEvent *) g_array_free(event_arr, FALSE);
  }
  song->ntempo = tempo_arr->len;
  song->tempos = (TempoChange *) g_array_free(tempo_arr, FALSE);
}

static unsigned vosfile_read_u8(FILE *file)
{
  int result = getc(file);

  g_assert(result != EOF);
  return (unsigned) result;
}

/* Little endian */
static unsigned vosfile_read_u16(FILE *file)
{
  guint16 tmp;
  int result;

  result = fread(&tmp, 2, 1, file); g_assert(result == 1);
  return GUINT16_FROM_LE(tmp);
}

static unsigned vosfile_read_u32(FILE *file)
{
  guint32 tmp;
  int result;

  result = fread(&tmp, 4, 1, file); g_assert(result == 1);
  return GUINT32_FROM_LE(tmp);
}

static char *vosfile_read_string(FILE *file)
{
  unsigned len;
  char *buf;
  int result;

  len = vosfile_read_u8(file); buf = g_new0(char, len + 1);
  result = fread(buf, 1, len, file); g_assert(result == (int) len);
  return buf;
}

static char *vosfile_read_string_fixed(FILE *file, unsigned len)
{
  char *buf;
  int result;
  
  buf = g_new0(char, len + 1);
  result = fread(buf, 1, len, file); g_assert(result == (int) len);
  return buf;
}

static void vosfile_read_info(VosSong *song, FILE *file, unsigned inf_ofs, unsigned inf_len)
{
  int result;
  unsigned inf_end_ofs = inf_ofs + inf_len, cur_ofs, next_ofs, key;
  char *title, *artist, *comment, *vos_author;
  unsigned song_type, ext_type, song_length, level;
  char buf[4];
  GArray *note_arrays_arr = g_array_new(FALSE, FALSE, sizeof(NoteArray));
  GArray *user_notes_arr = g_array_new(FALSE, FALSE, sizeof(UserNote));
  unsigned last_note_idx[NKEY], last_note_idx_long[NKEY], cur_un_idx;
  unsigned count[NKEY], count_long[NKEY];

  result = fseek(file, inf_ofs, SEEK_SET); g_assert(result == 0);
  /* Skip the "VOS1" header in e.g. test3.vos and test4.vos, if any */
  result = fread(buf, 4, 1, file);
  if (result == 1 && memcmp(buf, "VOS1", 4) == 0) { /* Found "VOS1" header */
    char *str1;
    result = fseek(file, 66, SEEK_CUR); g_assert(result == 0);
    str1 = vosfile_read_string(file); g_free(str1);
  } else {
    result = fseek(file, inf_ofs, SEEK_SET); g_assert(result == 0);
  }
  title = vosfile_read_string(file); g_free(title);
  artist = vosfile_read_string(file); g_free(artist);
  comment = vosfile_read_string(file); g_free(comment);
  vos_author = vosfile_read_string(file); g_free(vos_author);
  song_type = vosfile_read_u8(file); ext_type = vosfile_read_u8(file);
  song_length = vosfile_read_u32(file);
  level = vosfile_read_u8(file);

  result = fseek(file, 1023, SEEK_CUR); g_assert(result == 0);
  result = ftell(file); g_assert(result != -1); cur_ofs = result;
  for (key = 0; key < NKEY; key++) {
    count[key] = 0; count_long[key] = 0;
    last_note_idx[key] = IDX_NONE; last_note_idx_long[key] = IDX_NONE;
    song->first_un_idxs[key] = IDX_NONE; song->first_un_idxs_long[key] = IDX_NONE;
  }
  cur_un_idx = 0;
  while (1) {
    unsigned type, nnote, i;
    gboolean is_user_arr; /* Whether the current note array is the one to be played by the user. */
    char dummy2[14];
    NoteArray cur_note_arr;
    TempoIt it;

    result = ftell(file); g_assert(result != -1); cur_ofs = result;
    if (cur_ofs == inf_end_ofs) break;
    type = vosfile_read_u32(file); nnote = vosfile_read_u32(file);
    next_ofs = cur_ofs + nnote * 13 + 22; is_user_arr = (next_ofs == inf_end_ofs);
    result = fread(dummy2, 14, 1, file); g_assert(result == 1);

    memset(&cur_note_arr, 0, sizeof(cur_note_arr));
    tempoit_init(song, &it);
    if (! is_user_arr) { cur_note_arr.num_note = nnote; cur_note_arr.notes = g_new0(Note, nnote); }
    for (i = 0; i < nnote; i++) {
      unsigned time, len;
      unsigned char cmd, note_num, vol;
      unsigned flags;
      gboolean is_user_note;
      double tt_start, tt_end, rt_start, rt_end;
      
      time = vosfile_read_u32(file); len = vosfile_read_u32(file);
      cmd = vosfile_read_u8(file); note_num = vosfile_read_u8(file); vol = vosfile_read_u8(file);
      flags = vosfile_read_u16(file); is_user_note = ((flags & 0x80) != 0);

      tt_start = time / 768.0; tt_end = (time + len) / 768.0;
      rt_start = tempoit_tt_to_rt(song, &it, tt_start); rt_end = tempoit_tt_to_rt(song, &it, tt_end);
      song->max_rt = MAX(song->max_rt, rt_end);
      if (! is_user_arr) {
	Note *note = &cur_note_arr.notes[i];
	note->tt_start = tt_start; note->tt_end = tt_end;
	note->cmd = cmd; note->note_num = note_num; note->vol = vol;
	note->is_user = is_user_note;
      } else {
	UserNote un;
	
	g_assert(is_user_note);
	memset(&un, 0, sizeof(un));
	un.tt_start = tt_start; un.rt_start = rt_start;
	un.tt_end = tt_end; un.rt_end = rt_end;
	un.cmd = cmd; un.note_num = note_num; un.vol = vol;
	un.key = (flags & 0x70) >> 4; un.color = (flags & 0x0f); un.is_long = ((flags & 0x8000) != 0);
	un.tt_stop = (un.is_long) ? un.tt_end : un.tt_start;
	un.rt_stop = (un.is_long) ? un.rt_end : un.rt_start;
	g_assert(un.key < NKEY);
	un.prev_idx = last_note_idx[un.key]; un.next_idx = IDX_NONE;
	if (last_note_idx[un.key] != IDX_NONE) {
	  UserNote *last_un = &g_array_index(user_notes_arr, UserNote, last_note_idx[un.key]);
	  if (un.tt_start < last_un->tt_stop) { /* Notes on the same key should never overlap */
	    g_warning("User note %u ignored because of overlapping notes.", i);
	    goto skip_user_note;
	  }
	  last_un->next_idx = cur_un_idx;
	} else song->first_un_idxs[un.key] = cur_un_idx; /* First note on this key */
	last_note_idx[un.key] = cur_un_idx;
	un.count = ++count[un.key];
	if (un.is_long) {
	  un.prev_idx_long = last_note_idx_long[un.key]; un.next_idx_long = IDX_NONE;
	  if (last_note_idx_long[un.key] != IDX_NONE)
	    g_array_index(user_notes_arr, UserNote, last_note_idx_long[un.key]).next_idx_long = cur_un_idx;
	  else song->first_un_idxs_long[un.key] = cur_un_idx;
	  last_note_idx_long[un.key] = cur_un_idx;
	  un.count_long = ++count_long[un.key];
	} else { un.prev_idx_long = IDX_NONE; un.next_idx_long = IDX_NONE; }
	g_array_append_val(user_notes_arr, un); cur_un_idx++;
      skip_user_note: ;
      }
    }
    if (! is_user_arr) g_array_append_val(note_arrays_arr, cur_note_arr);
    cur_ofs = next_ofs; g_assert((unsigned) ftell(file) == cur_ofs);
  }
  song->num_note_array = note_arrays_arr->len;
  song->note_arrays = (NoteArray *) g_array_free(note_arrays_arr, FALSE);
  song->num_user_note = user_notes_arr->len;
  song->user_notes = (UserNote *) g_array_free(user_notes_arr, FALSE);
}

static VosSong *read_vos_file(const char *fname)
{
  VosSong *song;
  FILE *file;
  unsigned ofs, next_ofs, file_size;
  unsigned inf_ofs = 0, inf_len = 0, mid_ofs = 0, mid_len = 0;
  char *seg_name;
  unsigned magic;
  int result;
  
  file = fopen(fname, "rb"); g_assert(file != NULL);
  song = g_new0(VosSong, 1); song->max_rt = 0.0;
  hash_file(file, song->hash, SHA_DIGEST_LENGTH);
  result = fseek(file, 0, SEEK_END); g_assert(result == 0);
  file_size = ftell(file);
  result = fseek(file, 0, SEEK_SET); g_assert(result == 0);
  magic = vosfile_read_u32(file); g_assert(magic == 3);
  /* Read the segments */
  ofs = vosfile_read_u32(file);
  while (1) {
    seg_name = vosfile_read_string_fixed(file, 16);
    if (strcmp(seg_name, "EOF") == 0 || ofs == file_size) { g_free(seg_name); break; }
    next_ofs = vosfile_read_u32(file);
    if (strcmp(seg_name, "inf") == 0) { inf_ofs = ofs; inf_len = next_ofs - ofs; }
    else if (strcmp(seg_name, "mid") == 0) { mid_ofs = ofs; mid_len = next_ofs - ofs; }
    else g_assert_not_reached();
    g_free(seg_name);
    ofs = next_ofs;
  }
  g_assert(inf_len != 0); g_assert(mid_len != 0);

  vosfile_read_midi(song, file, mid_ofs, mid_len);
  vosfile_read_info(song, file, inf_ofs, inf_len);
  
  fclose(file);
  return song;
}

#define SET_COLOR(idx, r, g, b)					\
do {								\
  game->colors[(idx)].red = (guint16) ((double) (r) * 65535);	\
  game->colors[(idx)].green = (guint16) ((double) (g) * 65535);	\
  game->colors[(idx)].blue = (guint16) ((double) (b) * 65535);	\
 } while (0)

/* h is in 0..3 */
static void hsv2rgb(double h, double s, double v, double *pr, double *pg, double *pb)
{
  double x0 = (1.0 - s) * v, dx = s * v;
  double r, g, b;

  if (h < 1.0) { r = x0 + (1.0 - h) * dx; g = x0 + h * dx; b = x0; }
  else if (h < 2.0) { r = x0; g = x0 + (2.0 - h) * dx; b = x0 + (h - 1.0) * dx; }
  else { r = x0 + (h - 2.0) * dx; g = x0; b = x0 + (3.0 - h) * dx; }
  r = pow(r, 0.45); g = pow(g, 0.45); b = pow(b, 0.45);
  *pr = r; *pg = g; *pb = b;
}

static void init_colors(VosGame *game)
{
  gboolean *success = g_newa(gboolean, NCOLOR);
  unsigned i;

  game->colors = g_new0(GdkColor, NCOLOR);
  SET_COLOR(COLOR_BLACK, 0.0, 0.0, 0.0);
  SET_COLOR(COLOR_WHITE, 1.0, 1.0, 1.0);
  for (i = COLOR_KEYPRESS_START; i <= COLOR_KEYPRESS; i++) {
    double x = (double) (i - COLOR_KEYPRESS_START) / (COLOR_KEYPRESS - COLOR_KEYPRESS_START);
    double x0 = pow(x, 0.45), x1 = pow(1.0 - x, 0.45);
    SET_COLOR(i, 0.6 * x1, 0.0, 0.6 * x0);
  }
  g_assert(COLOR_NOTE_END == COLOR_NOTE_START + 31);
  for (i = 0; i < 16; i++) {
    unsigned j = (i * 5) % 16;
    double r, g, b;
    if (j == 0) r = g = b = 0.9;
    else {
      double h = (j - 1.0) * 0.2;
      hsv2rgb(h, 0.5, 1.0, &r, &g, &b);
    }
    SET_COLOR(COLOR_NOTE_START + i * 2, r, g, b);
    SET_COLOR(COLOR_NOTE_START + i * 2 + 1, r * 0.8, g * 0.8, b * 0.8);
  }
  g_assert(COLOR_MARKER_END == COLOR_MARKER_START + 31);
  for (i = 0; i < 32; i++) {
    double h = 1.0 + i * (3.0 / 32.0);
    double r, g, b;
    if (h >= 3.0) h -= 3.0;
    hsv2rgb(h, 1.0, 1.0, &r, &g, &b);
    SET_COLOR(COLOR_MARKER_START + i, r, g, b);
  }
  gdk_colormap_alloc_colors(gdk_colormap_get_system(), game->colors, NCOLOR, FALSE, TRUE, success);
}

static double get_rt(VosGame *game)
{
  GTimeVal cur_time;

  if (fast_mode) return game->start_rt;
  else {
    if (game->is_paused) cur_time = game->pause_time; else g_get_current_time(&cur_time);
    return game->start_rt + timeval_diff(&game->start_time, &cur_time);
  }
}

static void update_game_status(VosGame *game);
static gboolean on_refresh_timeout(gpointer data)
{
  VosGame *game = (VosGame *) data;
  GdkWindow *win = game->main_area->window;
  int width, height;
  GdkRectangle rect;
  double cur_rt = get_rt(game);

  gdk_drawable_get_size(GDK_DRAWABLE(win), &width, &height);
  rect.x = 0; rect.y = 0; rect.width = width; rect.height = height;
  gdk_window_invalidate_rect(win, &rect, FALSE);
  if (cur_rt - game->last_status_upd_rt >= 0.1) update_game_status(game);
  return TRUE; /* keep the timeout */
}

#define TRANSX(vx) ((int) floor(((double) (vx) - vx_min) * xstep + 0.5))
#define TRANSY(vy) ((int) floor(((double) (vy) - vy_min) * ystep + 0.5))

/* Note: GTK already provides double-buffering, so we need not do that */
static void do_redraw(VosGame *game)
{
  VosSong *song = game->song;
  GdkWindow *win = game->main_area->window;
  GdkDrawable *d = GDK_DRAWABLE(win);
  GdkGC *gc, *gc2;
  int width, height;
  double xstep, ystep;
  unsigned i;
  int y0, y;
  double cur_rt = get_rt(game);

  if (win == NULL) return;
  gc = gdk_gc_new(d); gc2 = gdk_gc_new(d);
  gdk_drawable_get_size(d, &width, &height); width = MAX(width, 1); height = MAX(height, 1);
  xstep = (width - 1) / (vx_max - vx_min); ystep = (height - 1) / (vy_max - vy_min);
  y0 = TRANSY(0.0);
  /* Clear */
  gdk_gc_set_foreground(gc, &game->colors[COLOR_BLACK]);
  gdk_draw_rectangle(d, gc, TRUE, 0, 0, width, height);
  /* Vertical bars */
  gdk_gc_set_foreground(gc, &game->colors[COLOR_WHITE]);
  for (i = 0; i <= NKEY; i++) {
    int x = TRANSX(i);
    gdk_draw_line(d, gc, x, 0, x, height - 1);
  }
  /* Keypresses */
  for (i = 0; i < NKEY; i++)
    if (game->keys[i].key_down) {
      int x0 = TRANSX(i), x1 = TRANSX(i + 1);
      int c = floor((cur_rt - game->keys[i].key_down_rt) * 20.0);
      unsigned color = COLOR_KEYPRESS_START + MIN(MAX(c, 0), COLOR_KEYPRESS - COLOR_KEYPRESS_START);
      gdk_gc_set_foreground(gc, &game->colors[color]);
      gdk_draw_rectangle(d, gc, TRUE, x0 + 1, 0, x1 - x0 - 1, height - 1);
    }
  /* Notes */
  for (i = 0; i < NKEY; i++) {
    int x0 = TRANSX(i + 0.1), x1 = TRANSX(i + 0.9);
    double tmin = cur_rt - vy_max / game->speed, tmax = cur_rt - vy_min / game->speed;
    unsigned idx = game->cur_un_idxs[i];
    UserNote *un;

    if (idx == IDX_NONE) continue; /* No notes on that key */
    while (un = &song->user_notes[idx], un->rt_stop >= tmin && un->prev_idx != IDX_NONE) idx = un->prev_idx;
    while (un = &song->user_notes[idx], un->rt_stop < tmin && un->next_idx != IDX_NONE) idx = un->next_idx;
    while (un = &song->user_notes[idx], un->rt_start < tmax) {
      double t0 = MAX(un->rt_start, tmin), t1 = MIN(un->rt_stop, tmax);
      unsigned color = COLOR_NOTE_START + un->color * 2;
      if (un->is_long && t0 < t1) { /* Stem */
	double vy0 = (cur_rt - t0) * game->speed, vy1 = (cur_rt - t1) * game->speed;
	int y0 = TRANSY(vy0), y1 = TRANSY(vy1);
	if (y1 < y0) {
	  int w = MIN((int) note_thickness, x1 - x0 + 1);
	  gdk_gc_set_foreground(gc, &game->colors[color]);
	  gdk_draw_rectangle(d, gc, TRUE, x0, y1, w, y0 - y1);
	  gdk_draw_rectangle(d, gc, TRUE, x1 - w + 1, y1, w, y0 - y1);
	  gdk_draw_line(d, gc, x0, y1, x0, y0 - 1);
	  gdk_draw_line(d, gc, x1, y1, x1, y0 - 1);
	}
      }
      if (un->rt_start >= tmin && un->rt_start < tmax) { /* Lower cap */
	double vy_start = (cur_rt - un->rt_start) * game->speed;
	int y_start = TRANSY(vy_start);
	gdk_gc_set_foreground(gc, &game->colors[color]);
	gdk_draw_rectangle(d, gc, TRUE, x0, y_start - note_thickness, x1 + 1 - x0, note_thickness);
      }
      if (un->is_long && un->rt_stop >= tmin && un->rt_stop < tmax) { /* Upper cap */
	double vy_stop = (cur_rt - un->rt_stop) * game->speed;
	int y_stop = TRANSY(vy_stop);
	gdk_gc_set_foreground(gc, &game->colors[color]);
	gdk_draw_rectangle(d, gc, TRUE, x0, y_stop - note_thickness, x1 + 1 - x0, note_thickness);
      }
      if (un->next_idx != IDX_NONE) idx = un->next_idx;
      else break;
    }
    game->cur_un_idxs[i] = idx;
  }
  /* Markers */
  gdk_gc_set_line_attributes(gc2, 5, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
  for (i = 0; i < NKEY; i++) {
    GQueue *markers = game->keys[i].markers;
    Marker *m;
    GList *p;
    static const double tscale = 3.0;
    int x0 = TRANSX(i + 0.5), y0 = TRANSY(0);

    while (m = (Marker *) g_queue_peek_head(markers), m != NULL && (cur_rt - m->rt) * tscale > 1.0) {
      g_free(m); g_queue_pop_head(markers); /* Pop overly old markers */
    }
    p = markers->head;
    while (p) {
      int n, r;
      double vr;
      
      m = (Marker *) p->data;
      /* NOTE: because m->rt is quantized and cur_rt is not, vr might be slightly negative */
      vr = (cur_rt - m->rt) * tscale * 0.8; vr = MAX(vr, 0.0); r = (int) floor(vr * xstep);
      n = (int) floor(m->dt / score_dt0 * 2 * 16 + 0.5);
      if (n < -16) n = -16; if (n > 16) n = 16; if (n < 0) n += 32;
      gdk_gc_set_foreground(gc2, &game->colors[COLOR_MARKER_START+n]);
      gdk_draw_arc(d, gc2, FALSE, x0 - r, y0 - r, 2 * r, 2 * r, 0, 64 * 360);
      p = p->next;
    }
  }
  /* Rotating markers of long notes */
  for (i = 0; i < NKEY; i++) {
    KeyState *ks = &game->keys[i];
    gboolean is_good = FALSE, is_bad = FALSE;
    if (ks->key_down) {
      if (ks->un_idx == IDX_NONE) is_bad = TRUE;
      else {
	UserNote *un = &game->song->user_notes[ks->un_idx];
	double rt_start = tempoit_tt_to_rt(game->song, &game->it, un->tt_start);
	double rt_end = tempoit_tt_to_rt(game->song, &game->it, un->tt_end);
	if (cur_rt >= rt_start && cur_rt < rt_end) { if (un->is_long) is_good = TRUE; }
	else is_bad = TRUE;
      }
    }
    if (is_good || is_bad) {
      unsigned color = (is_good) ? COLOR_MARKER_BEST : COLOR_MARKER_WORST;
      double angle = (cur_rt - floor(cur_rt)) * (2.0 * G_PI);
      int x0 = TRANSX(i + 0.5), y0 = TRANSY(0);
      int dx = ceil(0.4 * xstep * cos(angle)), dy = ceil(0.4 * xstep * sin(angle));
      gdk_gc_set_foreground(gc2, &game->colors[color]);
      gdk_draw_line(d, gc2, x0 - dx, y0 - dy, x0 + dx, y0 + dy);
      gdk_draw_line(d, gc2, x0 - dy, y0 + dx, x0 + dy, y0 - dx);
    }
  }
  /* Horizontal bars */
  gdk_gc_set_foreground(gc, &game->colors[COLOR_WHITE]);
  y = y0; gdk_draw_line(d, gc, 0, y, width - 1, y); /* Lower line */
  y = y0 - note_thickness - 1; gdk_draw_line(d, gc, 0, y, width - 1, y); /* Upper line */
  g_object_unref(G_OBJECT(gc)); g_object_unref(G_OBJECT(gc2));
}

static gboolean on_main_area_expose(GtkWidget *widget, GdkEventExpose *ev, gpointer user_data)
{
  VosGame *game = (VosGame *) user_data;

  do_redraw(game);
  return TRUE; /* End event handling */
}

static gboolean on_main_window_delete(GtkWidget *widget, GdkEvent *ev, gpointer user_data)
{
  gtk_main_quit();
  return TRUE; /* End event handling */
}

static gboolean on_event_timeout(gpointer data);
static void process_key(VosGame *game, double cur_rt, unsigned key, gboolean is_keypress);
static void update_game_score(VosGame *game, double cur_rt);
static void get_game_score(VosGame *game, GameScore *out);
/* Process the events in the scheduler whose time is up, and set the timeout for the next one */
static void process_events(VosGame *game)
{
  SchedEvent *ev;
  double cur_rt;
  gboolean is_quit = FALSE;

 restart:
  cur_rt = get_rt(game);
  while ((ev = get_first_event(game->sched)) != NULL) {
    if (cur_rt >= ev->rt) {
      snd_seq_event_t sev;
      gboolean no_output = FALSE;

      /*      g_message("Event cur_rt=%0.3f rt=%0.3f type=%d src=%u idx=%u",
	      cur_rt, ev->rt, ev->type, ev->src, ev->idx); */
      snd_seq_ev_clear(&sev);
      snd_seq_ev_set_source(&sev, game->seq_port_id);
      snd_seq_ev_set_dest(&sev, game->dest_client_id, game->dest_port_id);
      snd_seq_ev_set_direct(&sev);
      if (ev->type == SchedEventMidi) {
	MidiTrack *track = &game->song->tracks[ev->src];
	MidiEvent *mev = &track->events[ev->idx];
	unsigned cmd_type = (mev->cmd & 0x70) >> 4, chan = mev->cmd & 0x0f;;
	int pitch;
	switch (cmd_type) {
	case 0: snd_seq_ev_set_noteoff(&sev, chan, mev->a, mev->b); break; /* Note off */
	case 1: snd_seq_ev_set_noteon(&sev, chan, mev->a, mev->b); break; /* Note on */
	case 2: snd_seq_ev_set_keypress(&sev, chan, mev->a, mev->b); break; /* Key pressure */
	case 3: snd_seq_ev_set_controller(&sev, chan, mev->a, mev->b); break; /* Controller */
	case 4: snd_seq_ev_set_pgmchange(&sev, chan, mev->a); break; /* Program change */
	case 5: snd_seq_ev_set_chanpress(&sev, chan, mev->a); break; /* Channel pressure */
	case 6: /* Pitch wheel */
	  pitch = ((int) (((unsigned) (mev->b & 0x80) << 7) | (unsigned) (mev->a & 0x80))) - 0x2000;
	  snd_seq_ev_set_pitchbend(&sev, chan, pitch);
	  break;
	case 7: g_assert_not_reached(); break;
	}
	if (cmd_type == 3 && mev->a == 64) no_output = TRUE; /* Hold Pedal doesn't sound good... */
	if (ev->idx + 1 < track->nevent) {
	  MidiEvent *mev = &track->events[ev->idx+1];
	  double rt = tempoit_tt_to_rt(game->song, &game->it, mev->tt);
	  sched_event(game->sched, rt, SchedEventMidi, ev->src, ev->idx + 1);
	}
      } else if (ev->type == SchedEventNoteOn || ev->type == SchedEventNoteOff) {
	NoteArray *arr = &game->song->note_arrays[ev->src];
	Note *note = &arr->notes[ev->idx];
	unsigned chan = (note->cmd & 0x0f);
	unsigned char vol = (note->vol + 1) / 2; /* Change the volume for BGM */

	if (note->is_user) no_output = TRUE;
	if (ev->type == SchedEventNoteOn) {
	  double rt_end = tempoit_tt_to_rt(game->song, &game->it, note->tt_end);

	  snd_seq_ev_set_noteon(&sev, chan, note->note_num, vol);
	  sched_event(game->sched, rt_end, SchedEventNoteOff, ev->src, ev->idx);
	  if (ev->idx + 1 < arr->num_note) {
	    Note *note = &arr->notes[ev->idx+1];
	    double rt = tempoit_tt_to_rt(game->song, &game->it, note->tt_start);
	    sched_event(game->sched, rt, SchedEventNoteOn, ev->src, ev->idx + 1);
	  }
	} else snd_seq_ev_set_noteoff(&sev, chan, note->note_num, vol);
      } else if (ev->type == SchedEventUserNoteOff) { /* Delayed user note off */
	UserNote *un = &game->song->user_notes[ev->idx];
	snd_seq_ev_set_noteoff(&sev, un->cmd & 0x0f, un->note_num, un->vol);
      } else if (ev->type == SchedEventDemo) {
	InputEvent *iev = &game->in_evs[ev->idx];
	process_key(game, iev->rt, iev->key, iev->is_keypress);
	if (ev->idx + 1 < game->in_nev) {
	  InputEvent *iev = &game->in_evs[ev->idx+1];
	  sched_event(game->sched, iev->rt, SchedEventDemo, 0, ev->idx + 1);
	}
      } else if (ev->type == SchedEventRefreshScore) {
	GameScore sc, last_sc;
	char *text;
	unsigned i;
	double time = recent_scores_nperiod * score_refresh_period;
	double rate, total_rate;
	
	update_game_score(game, ev->rt); /* Which calls refresh_scores() */
	get_game_score(game, &sc);
	last_sc = game->recent_scores[0];
	for (i = 1; i < recent_scores_nperiod; i++) game->recent_scores[i-1] = game->recent_scores[i];
	game->recent_scores[recent_scores_nperiod-1] = sc;
	rate = (sc.score - last_sc.score) / time; total_rate = (sc.total_score - last_sc.total_score) / time;
	text = g_strdup_printf("Rate: %0.2f / %0.2f (%0.2f%%)", rate, total_rate,
			       (total_rate == 0.0) ? 100.0 : (rate / total_rate * 100.0));
	gtk_label_set_text(game->rate_label, text);
	g_free(text);
	if (ev->rt >= game->song->max_rt + 1.0) is_quit = TRUE;
	else sched_event(game->sched, ev->rt + score_refresh_period, SchedEventRefreshScore, 0, 0);
      } else if (ev->type == SchedEventAutoOn || ev->type == SchedEventAutoOff) {
	UserNote *un = &game->song->user_notes[ev->idx];
	double rt_end = un->rt_end;
	if (un->next_idx != IDX_NONE) rt_end = MIN(rt_end, game->song->user_notes[un->next_idx].rt_start);
	if (! un->is_long) rt_end = MIN(rt_end, un->rt_start + 0.1);
	if (ev->type == SchedEventAutoOn) {
	  process_key(game, un->rt_start, un->key, TRUE);
	  sched_event(game->sched, rt_end, SchedEventAutoOff, ev->src, ev->idx);
	} else {
	  process_key(game, rt_end, un->key, FALSE);
	  if (un->next_idx != IDX_NONE)
	    sched_event(game->sched, game->song->user_notes[un->next_idx].rt_start, SchedEventAutoOn,
			ev->src, un->next_idx);
	}
      } else g_assert_not_reached();
      if (! no_output && ! fast_mode) snd_seq_event_output(game->seq, &sev);
      remove_event(game->sched, ev);
    } else break;
  }
  snd_seq_drain_output(game->seq);
  if (! fast_mode) {
    if (ev == NULL) game->event_timeout_id = 0;
    else {
      int dt = (int) ceil((ev->rt - cur_rt) * 1000.0);
      g_assert(dt > 0);
      /*    g_message("Schedule: dt=%d ms", dt); */
      game->event_timeout_id = gtk_timeout_add(dt, on_event_timeout, game);
    }
    if (is_quit) gtk_main_quit();
  } else {
    if (is_quit || ev == NULL) return;
    else { game->start_rt = ev->rt; goto restart; }
  }
}

static gboolean on_event_timeout(gpointer data)
{
  VosGame *game = (VosGame *) data;

  process_events(game);
  return FALSE; /* remove the timeout */
}

static void add_marker(VosGame *game, unsigned key, double rt, double dt)
{
  Marker *m;

  m = g_new0(Marker, 1); m->rt = rt; m->dt = dt;
  g_queue_push_tail(game->keys[key].markers, m);
}

/* Update the state of the comparer after an event at cur_rt.
   If is_dummy is true, just update the tables with the new cur_rt, but there is no real event. */
#define PREV_IDX(un) ((is_keypress) ? (un)->prev_idx : (un)->prev_idx_long)
#define NEXT_IDX(un) ((is_keypress) ? (un)->next_idx : (un)->next_idx_long)
#define RT(un) ((is_keypress) ? (un)->rt_start : (un)->rt_end)
#define COUNT(un) ((is_keypress) ? (un)->count : (un)->count_long)
static void comparer_add_event(VosGame *game, double cur_rt, unsigned key, gboolean is_keypress,
			       gboolean is_dummy)
{
  KeyState *ks = &game->keys[key];
  Comparer *cmp = (is_keypress) ? &ks->cmp_press : &ks->cmp_release;
  double min_score = (is_keypress) ? -1.0 : 0.0, max_dt = (1.0 - min_score) * 0.5 * score_dt0;
  double min_rt = cur_rt - max_dt, max_rt = cur_rt + max_dt;
  unsigned idx = (is_keypress) ? game->cur_un_idxs[key] : game->cur_un_idxs_long[key];
  unsigned prev_idx, next_idx;
  UserNote *un;
  double un_rt;
  ComparerEntry ent;
  unsigned i0, i1;
  GArray *new_ents_arr;
  double score0, last_score, score_i0, score_i1;
  unsigned count0;

  if (idx == IDX_NONE) { /* No (long) note on this key */
    g_assert(cmp->nent == 0); if (! is_dummy) cmp->score0 += min_score;
  } else {
    new_ents_arr = g_array_new(FALSE, FALSE, sizeof(ComparerEntry));
    while (un = &game->song->user_notes[idx], prev_idx = PREV_IDX(un),
	   prev_idx != IDX_NONE && RT(&game->song->user_notes[prev_idx]) >= min_rt)
      idx = prev_idx;
    while (un = &game->song->user_notes[idx], next_idx = NEXT_IDX(un),
	   RT(un) < min_rt && next_idx != IDX_NONE)
      idx = next_idx;
    count0 = COUNT(un) - 1; if (RT(un) < min_rt) count0++;
    i0 = i1 = 0; score_i0 = score_i1 = cmp->score0;
    while (i1 < cmp->nent && RT(&game->song->user_notes[cmp->ents[i1].last_idx]) < min_rt) {
      score_i1 = cmp->ents[i1].score; i1++;
    }
    score0 = (is_dummy) ? score_i1 : score_i1 + min_score; last_score = score0;
    while (un = &game->song->user_notes[idx], un_rt = RT(un),
	   un_rt >= min_rt && un_rt <= max_rt) {
      double note_score = 1.0 - 2.0 * fabs(un_rt - cur_rt) / score_dt0;
      double score_match, score_ignore;
      note_score = MAX(note_score, min_score);
      while (i0 < cmp->nent && cmp->ents[i0].last_idx < idx) { score_i0 = cmp->ents[i0].score; i0++; }
      while (i1 < cmp->nent && cmp->ents[i1].last_idx <= idx) { score_i1 = cmp->ents[i1].score; i1++; }
      score_match = (is_dummy) ? score_i0 : score_i0 + note_score; /* Match the keypress with un */
      score_ignore = (is_dummy) ? score_i1 : score_i1 + min_score; /* Ignore the keypress */
      ent.last_idx = idx; ent.score = MAX(score_match, score_ignore);
      if (ent.score > last_score) { last_score = ent.score; g_array_append_val(new_ents_arr, ent); }
      next_idx = NEXT_IDX(un); if (next_idx != IDX_NONE) idx = next_idx; else break;
    }
    if (is_keypress) game->cur_un_idxs[key] = idx; else game->cur_un_idxs_long[key] = idx;
    g_free(cmp->ents); cmp->count0 = count0; cmp->score0 = score0; cmp->nent = new_ents_arr->len;
    cmp->ents = (ComparerEntry *) g_array_free(new_ents_arr, FALSE);
  }
}
#undef PREV_IDX
#undef NEXT_IDX
#undef RT
#undef COUNT

/* Count the long_score in the duration between long_rt and cur_rt, and update long_rt to cur_rt */
static void calc_long_score(VosGame *game, double cur_rt, unsigned key, gboolean key_down)
{
  UserNote *un;
  unsigned idx = game->cur_un_idxs[key];
  KeyState *ks = &game->keys[key];
  double rt0, rt1;
  double cur_time, okay_time = 0.0, scored_time = 0.0, total_time = cur_rt - ks->long_rt;

  if (total_time < 0.0) return; /* May happen if the user pressed some key during the wait time, or
				   due to small timing errors. */
  if (idx == IDX_NONE) { ks->long_rt = cur_rt; return; }
  while (un = &game->song->user_notes[idx], un->rt_start >= ks->long_rt && un->prev_idx != IDX_NONE)
    idx = un->prev_idx;
  while (1) {
    un = &game->song->user_notes[idx];
    rt0 = un->rt_start; rt1 = un->rt_end;
    if (un->next_idx != IDX_NONE) rt1 = MIN(rt1, game->song->user_notes[un->next_idx].rt_start);
    if (rt0 > cur_rt) break;
    cur_time = MIN(rt1, cur_rt) - MAX(rt0, ks->long_rt); cur_time = MAX(cur_time, 0.0);
    okay_time += cur_time; if (un->is_long) scored_time += cur_time;
    if (un->next_idx == IDX_NONE) break; else idx = un->next_idx;
  }
  g_assert(okay_time >= 0.0); g_assert(okay_time <= total_time + 1e-6);
  if (key_down) ks->long_score += scored_time - (total_time - okay_time);
  ks->total_long_score += scored_time; ks->long_rt = cur_rt;
  game->cur_un_idxs[key] = idx;
}

static void refresh_scores(VosGame *game, double cur_rt)
{
  unsigned key;

  for (key = 0; key < NKEY; key++) {
    KeyState *ks = &game->keys[key];
    comparer_add_event(game, cur_rt, key, TRUE, TRUE); comparer_add_event(game, cur_rt, key, FALSE, TRUE);
    calc_long_score(game, cur_rt, key, ks->key_down);
  }
}

/* cur_rt does not have to be the current time, but it must be monotonic */
static void process_key(VosGame *game, double cur_rt, unsigned key, gboolean is_keypress)
{
  KeyState *ks = &game->keys[key];
  snd_seq_event_t sev;

  /* NOTE: cur_rt is not quantized which may cause some minor errors in demo recording.  However,
     quantizing cur_rt may cause the scoring process to lose time-ordering, which can be very
     troublesome. */
  if (is_keypress && ks->key_down) return; /* Probably due to auto-repeat */
  if (! is_keypress && ! ks->key_down) { g_warning("Unexpected release of key %u", key); return; }
  ks->key_down = is_keypress; if (is_keypress) ks->key_down_rt = cur_rt;

  if (game->mode == VosGameModeNormal) { /* Record the keypress */
    InputEvent ev;
    ev.rt = cur_rt; ev.is_keypress = is_keypress; ev.key = key;
    g_array_append_val(game->input_events_arr, ev);
  }

  comparer_add_event(game, cur_rt, key, is_keypress, FALSE);
  calc_long_score(game, cur_rt, key, is_keypress ? FALSE : TRUE);
  update_game_score(game, cur_rt);
  
  snd_seq_ev_clear(&sev);
  snd_seq_ev_set_source(&sev, game->seq_port_id);
  snd_seq_ev_set_dest(&sev, game->dest_client_id, game->dest_port_id);
  snd_seq_ev_set_direct(&sev);
  if (is_keypress) { /* Need to find the closest UserNote */
    unsigned idx = game->cur_un_idxs[key], best_idx;
    double dt, best_dt;
    UserNote *un;

    if (idx == IDX_NONE) { ks->un_idx = IDX_NONE; add_marker(game, key, cur_rt, 100.0); }
    else {
      while (un = &game->song->user_notes[idx], un->rt_start >= cur_rt && un->prev_idx != IDX_NONE)
	idx = un->prev_idx; /* Go back to some note before the cur_rt, or the first note */
      best_idx = IDX_NONE; best_dt = 100.0;
      while (1) {
	dt = un->rt_start - cur_rt;
	/* Key pressed in the middle of a long note should be regarded as an attempt to finish it */
	if (un->is_long && dt <= 0.0 && un->rt_end - cur_rt >= 0.0) dt = 0.0;
	if (best_idx == IDX_NONE || fabs(dt) < fabs(best_dt)) { best_idx = idx; best_dt = dt; }
	if (un->rt_start >= cur_rt) break;
	if (un->next_idx == IDX_NONE) break;
	idx = un->next_idx; un = &game->song->user_notes[idx];
      }
      game->cur_un_idxs[key] = idx;
      ks->un_idx = idx = best_idx; un = &game->song->user_notes[idx];
      snd_seq_ev_set_noteon(&sev, un->cmd & 0x0f, un->note_num, un->vol);
      if (! fast_mode) snd_seq_event_output(game->seq, &sev);
      snd_seq_drain_output(game->seq);
      /* NOTE: for notes pressed in the middle of long notes, dt above is zero, but for scoring purposes
	 some penalty is needed, thus we use un->rt_start - cur_rt here */
      add_marker(game, key, cur_rt, un->rt_start - cur_rt);
    }
  } else {
    if (ks->un_idx != IDX_NONE) {
      unsigned idx = ks->un_idx;
      UserNote *un = &game->song->user_notes[idx];
      ks->un_idx = IDX_NONE;
      /* Automatically lengthen short notes if needed */
      if (short_note_auto_len && ! un->is_long && cur_rt < un->rt_end)
	sched_event(game->sched, un->rt_end, SchedEventUserNoteOff, 0, idx);
      else {
	snd_seq_ev_set_noteoff(&sev, un->cmd & 0x0f, un->note_num, un->vol);
	if (! fast_mode) snd_seq_event_output(game->seq, &sev);
	snd_seq_drain_output(game->seq);
      }
      if (un->is_long) add_marker(game, key, cur_rt, un->rt_end - cur_rt);
    }
  }
}

static void update_game_state(VosGame *game)
{
  gtk_label_set_text(game->state_label, (game->is_paused) ? "Paused" : "Running");
}

/* Things like time etc. */
static void update_game_status(VosGame *game)
{
  char *text;
  double cur_rt = get_rt(game);
  unsigned cur_rt_nsec = (unsigned) floor(MAX(cur_rt, 0.0));
  unsigned total_nsec = (unsigned) floor(game->song->max_rt);

  game->last_status_upd_rt = cur_rt;
  text = g_strdup_printf("Time: %u:%02u / %u:%02u", cur_rt_nsec / 60, cur_rt_nsec % 60,
			 total_nsec / 60, total_nsec % 60);
  gtk_label_set_text(game->time_label, text);
  g_free(text);
}

static void get_game_score(VosGame *game, GameScore *out)
{
  double score = 0.0, total_score = 0.0;
  unsigned key;

  for (key = 0; key < NKEY; key++) {
    KeyState *ks = &game->keys[key];
    Comparer *cmp;
    cmp = &ks->cmp_press;
    if (cmp->nent == 0) { score += cmp->score0; total_score += cmp->count0; }
    else {
      ComparerEntry *ent = &cmp->ents[cmp->nent-1];
      score += ent->score; total_score += game->song->user_notes[ent->last_idx].count;
    }
    cmp = &ks->cmp_release;
    if (cmp->nent == 0) { score += cmp->score0; total_score += cmp->count0; }
    else {
      ComparerEntry *ent = &cmp->ents[cmp->nent-1];
      score += ent->score; total_score += game->song->user_notes[ent->last_idx].count_long;
    }
    score += ks->long_score; total_score += ks->total_long_score;
  }
  out->score = score; out->total_score = total_score;
}

static void update_game_score(VosGame *game, double cur_rt)
{
  GameScore sc;
  char *text;

  refresh_scores(game, cur_rt);
  get_game_score(game, &sc);
  text = g_strdup_printf("Score: %0.2f / %0.2f (%0.2f%%)", sc.score, sc.total_score,
			 (sc.total_score == 0.0) ? 100.0 : (sc.score / sc.total_score * 100.0));
  gtk_label_set_text(game->score_label, text);
  g_free(text);
}

static void pause_game(VosGame *game)
{
  g_assert(! game->is_paused);
  game->is_paused = TRUE;
  g_get_current_time(&game->pause_time);
  if (game->event_timeout_id) { gtk_timeout_remove(game->event_timeout_id); game->event_timeout_id = 0; }
  on_refresh_timeout(game);
  if (game->refresh_timeout_id) {
    gtk_timeout_remove(game->refresh_timeout_id); game->refresh_timeout_id = 0;
  }
  update_game_state(game); update_game_status(game);
}

static void unpause_game(VosGame *game)
{
  g_assert(game->is_paused);
  game->start_rt = get_rt(game);
  game->is_paused = FALSE;
  g_get_current_time(&game->start_time);
  process_events(game);
  on_refresh_timeout(game);
  game->refresh_timeout_id = gtk_timeout_add(5, on_refresh_timeout, game); /* 200fps */
  update_game_state(game); update_game_status(game);
}

static gboolean on_main_area_key(GtkWidget *widget, GdkEventKey *ev, gpointer user_data)
{
  VosGame *game = (VosGame *) user_data;
  double cur_rt = get_rt(game);
  unsigned key = IDX_NONE;
  gboolean is_keypress = (ev->type == GDK_KEY_PRESS);

  switch (ev->keyval) {
  case GDK_z: key = 0; break;
  case GDK_x: key = 1; break;
  case GDK_c: key = 2; break;
  case GDK_v: key = 3; break;
  case GDK_comma: key = 4; break;
  case GDK_period: key = 5; break;
  case GDK_slash: key = 6; break;
  case GDK_space:
    if (is_keypress) {
      if (! game->is_paused) pause_game(game);
      else unpause_game(game);
    }
    break;
  default: return FALSE;
  }
  if (key != IDX_NONE && game->mode == VosGameModeNormal && ! game->is_paused)
    process_key(game, cur_rt, key, is_keypress);
  return TRUE; /* event handled */
}

static guint32 pev_read_u32(FILE *file)
{
  guint32 tmp;
  int result;
  result = fread(&tmp, 4, 1, file); g_assert(result == 1);
  return GUINT32_FROM_LE(tmp);
}

static void pev_write_u32(FILE *file, guint32 value)
{
  guint32 tmp = GUINT32_TO_LE(value);
  int result;
  result = fwrite(&tmp, 4, 1, file); g_assert(result == 1);
}

static double pev_read_double(FILE *file)
{
  union { guint64 u64; double dbl; } u;
  guint64 tmp;
  int result;
  
  result = fread(&tmp, 8, 1, file); g_assert(result == 1);
  u.u64 = GUINT64_FROM_LE(tmp);
  return u.dbl;
}

static void pev_write_double(FILE *file, double value)
{
  union { guint64 u64; double dbl; } u;
  guint64 tmp;
  int result;
  
  u.dbl = value;
  tmp = GUINT64_TO_LE(u.u64);
  result = fwrite(&tmp, 8, 1, file); g_assert(result == 1);
}

static unsigned pev_read_vl(FILE *file)
{
  unsigned value = 0, nshift = 0;
  int result;
  unsigned char ch;
  while (1) {
    result = getc(file); g_assert(result != EOF);
    ch = (unsigned) result;
    value |= (ch & 0x7f) << nshift; nshift += 7;
    if (! (ch & 0x80)) break;
  }
  return value;
}

static void pev_write_vl(FILE *file, unsigned value)
{
  unsigned char ch;
  int result;
  do {
    ch = value & 0x7f; value >>= 7;
    if (value != 0) ch |= 0x80;
    result = putc(ch, file); g_assert(result != EOF);
  } while (value != 0);
}

/* NOTE: the length field is already read */
static char *pev_read_str(FILE *file, unsigned len)
{
  char *str;
  int result;

  str = g_new0(char, len + 1);
  result = fread(str, 1, len, file); g_assert(result == (int) len);
  return str;
}

/* This also writes the length field.  NOTE: the written string is not NULL-terminated */
static void pev_write_str(FILE *file, const char *str)
{
  unsigned len = strlen(str);
  int result;
  pev_write_vl(file, len);
  result = fwrite(str, 1, len, file); g_assert(result == (int) len);
}

static void load_input_events(const char *in_fname, GArray *events_arr, unsigned char *hash,
			      gboolean *has_hash)
{
  FILE *file;
  unsigned magic, type, len;
  time_t time;
  char *user_name;
  int irt;
  int result;

  *has_hash = FALSE;
  file = fopen(in_fname, "rb"); g_assert(file != NULL);
  magic = pev_read_u32(file); g_assert(magic == PEV_MAGIC);
  tempo_factor = 1.0;
  while (1) {
    type = pev_read_vl(file); if (type == PEV_HEADER_END) break;
    len = pev_read_vl(file);
    switch (type) {
    case PEV_HASH_SHA1:
      g_assert(len == SHA_DIGEST_LENGTH);
      *has_hash = TRUE;
      result = fread(hash, SHA_DIGEST_LENGTH, 1, file); g_assert(result == 1);
      break;
    case PEV_TEMPO_FACTOR: g_assert(len == 8); tempo_factor = pev_read_double(file); break;
    case PEV_SPEED: g_assert(len == 8); default_speed = pev_read_double(file); break;
    case PEV_TIME_UNIT: g_assert(len == 8); default_pev_time_unit = pev_read_double(file); break;
    case PEV_DATE:
      g_assert(len == 4); time = pev_read_u32(file);
      g_message("Demo recorded at %s", ctime(&time));
      break;
    case PEV_USER_NAME:
      user_name = pev_read_str(file, len);
      g_message("Recorded by: %s", user_name);
      g_free(user_name);
      break;
    default:
      g_warning("Unknown PEV header type %u", type);
      result = fseek(file, len, SEEK_CUR); g_assert(result == 0);
      break;
    }
  }
  irt = 0;
  while (1) {
    int idt;
    InputEvent ev;
    
    type = pev_read_vl(file);
    if (type == PEV_EV_END) break;
    else if (type >= PEV_EV_PRESS && type < PEV_EV_PRESS + NKEY) {
      ev.key = type - PEV_EV_PRESS; ev.is_keypress = TRUE;
    } else if (type >= PEV_EV_RELEASE && type < PEV_EV_RELEASE + NKEY) {
      ev.key = type - PEV_EV_RELEASE; ev.is_keypress = FALSE;
    } else g_assert_not_reached();
    idt = pev_read_vl(file); g_assert(idt >= 0); irt += idt; ev.rt = irt * default_pev_time_unit;
    g_array_append_val(events_arr, ev);
    /*    g_message("event: pos=%u time=%0.3f", (unsigned) ftell(file), ev.rt); */
  }
  fclose(file);
}

static void save_input_events(VosGame *game, const char *out_fname)
{
  FILE *file;
  unsigned i, nev = game->input_events_arr->len;
  InputEvent *evs = (InputEvent *) g_array_free(game->input_events_arr, FALSE);
  int last_irt;
  int result;

  if (game->mode != VosGameModeNormal) goto no_save;
  if (out_fname != NULL) file = fopen(out_fname, "wb");
  else {
    char *fname;
    GError *err = NULL;
    int fd = g_file_open_tmp("pmg-save-XXXXXX", &fname, &err);
    if (fd == -1) {
      g_warning("Cannot write autosave'd event list file: %s.", err->message);
      g_error_free(err);
      goto no_save;
    }
    g_message("Autosave file name: %s", fname);
    g_free(fname);
    file = fdopen(fd, "wb");
  }
  if (file == NULL) { g_warning("Cannot write event list file %s.", out_fname); goto no_save; }
  /* Write file header */
  pev_write_u32(file, PEV_MAGIC);
  pev_write_vl(file, PEV_HASH_SHA1); pev_write_vl(file, SHA_DIGEST_LENGTH);
  result = fwrite(game->song->hash, SHA_DIGEST_LENGTH, 1, file); g_assert(result == 1);
  if (tempo_factor != 1.0) {
    pev_write_vl(file, PEV_TEMPO_FACTOR); pev_write_vl(file, 8); pev_write_double(file, tempo_factor);
  }
  pev_write_vl(file, PEV_SPEED); pev_write_vl(file, 8); pev_write_double(file, game->speed);
  pev_write_vl(file, PEV_TIME_UNIT); pev_write_vl(file, 8); pev_write_double(file, game->pev_time_unit);
  pev_write_vl(file, PEV_DATE); pev_write_vl(file, 4); pev_write_u32(file, game->start_time.tv_sec);
  pev_write_vl(file, PEV_USER_NAME); pev_write_str(file, game->user_name);
  pev_write_vl(file, PEV_HEADER_END);
  last_irt = 0;
  for (i = 0; i < nev; i++) {
    InputEvent *ev = &evs[i];
    int irt = (int) floor(ev->rt / game->pev_time_unit + 0.5);
    int idt = irt - last_irt;
    g_assert(idt >= 0);
    pev_write_vl(file, ((ev->is_keypress) ? PEV_EV_PRESS : PEV_EV_RELEASE) + ev->key);
    pev_write_vl(file, idt);
    last_irt = irt;
  }
  pev_write_vl(file, PEV_EV_END);
  fclose(file);
 no_save:
  g_free(evs);
}

int main(int argc, char **argv)
{
  VosGameMode mode = VosGameModeNormal;
  const char *vos_fname;
  poptContext popt_ctx;
  InputEvent *in_evs = NULL;
  unsigned in_nev = 0;
  unsigned char demo_song_hash[SHA_DIGEST_LENGTH];
  gboolean demo_song_has_hash = FALSE;
  VosGame *game;
  GladeXML *gx;
  unsigned key;
  GtkWidget *main_win;
  unsigned i;
  int result;
  PangoAttrList *attr_list;
  PangoAttribute *attr;
  GameScore sc;
  
  gtk_init(&argc, &argv);
  glade_init();

  /* Parse command-line options */
  popt_ctx = poptGetContext("pmg", argc, (const char **) argv, popt_table, 0);
  while ((result = poptGetNextOpt(popt_ctx)) > 0) {
    if (result == 'i') {
      GArray *in_arr = g_array_new(FALSE, FALSE, sizeof(InputEvent));
      load_input_events(pev_in_fname, in_arr, demo_song_hash, &demo_song_has_hash);
      in_nev = in_arr->len; in_evs = (InputEvent *) g_array_free(in_arr, FALSE);
      mode = VosGameModePlayDemo;
    } else if (result == 'A') mode = VosGameModeAuto;
    else g_assert_not_reached();
  }
  g_assert(result < 0);
  if (result < -1) {
    fprintf(stderr, "%s: %s\n", poptBadOption(popt_ctx, POPT_BADOPTION_NOALIAS), poptStrerror(result));
    exit(1);
  }
  vos_fname = poptGetArg(popt_ctx);
  if (vos_fname == NULL) { poptPrintUsage(popt_ctx, stderr, 0); exit(1); }

  game = g_new0(VosGame, 1); game->mode = mode; game->user_name = g_get_user_name();
  game->in_evs = in_evs; game->in_nev = in_nev;

  /* Initialize sound */
  result = snd_seq_open(&game->seq, "default", SND_SEQ_OPEN_OUTPUT, SND_SEQ_NONBLOCK);
  g_assert(result == 0);
  game->seq_client_id = snd_seq_client_id(game->seq);
  game->seq_port_id = snd_seq_create_simple_port(game->seq, "PMG output", SND_SEQ_PORT_CAP_READ,
						 SND_SEQ_PORT_TYPE_MIDI_GENERIC);
  g_assert(game->seq_port_id >= 0);
  result = sscanf(midi_port_spec, "%u:%u", &game->dest_client_id, &game->dest_port_id);
  g_assert(result == 2);
  g_message("Using MIDI sequencer port: %u:%u", game->dest_client_id, game->dest_port_id);
  result = snd_seq_connect_to(game->seq, game->seq_port_id, game->dest_client_id, game->dest_port_id);
  g_assert(result == 0);
  
  /* Read song */
  game->song = read_vos_file(vos_fname);
  if (game->mode == VosGameModePlayDemo && demo_song_has_hash)
    if (memcmp(demo_song_hash, game->song->hash, SHA_DIGEST_LENGTH) != 0)
      g_warning("The SHA-1 hashes do not match.  The demo is probably recorded with a different song.");

  /* Build the user interface */
  gx = glade_xml_new(glade_fname, "main_window", NULL); g_assert(gx != NULL);
  glade_xml_signal_autoconnect(gx);

  /* Initialize the game */
  GET_WIDGET(gx, main_win, "main_window", GTK_WIDGET);
  GET_WIDGET(gx, game->main_area, "main_area", GTK_WIDGET);
  GET_WIDGET(gx, game->state_label, "state_label", GTK_LABEL);
  GET_WIDGET(gx, game->time_label, "time_label", GTK_LABEL);
  GET_WIDGET(gx, game->score_label, "score_label", GTK_LABEL);
  GET_WIDGET(gx, game->rate_label, "rate_label", GTK_LABEL);
  attr_list = pango_attr_list_new();
  attr = pango_attr_family_new("sans-serif"); attr->start_index = 0; attr->end_index = G_MAXINT;
  pango_attr_list_insert(attr_list, attr);
  attr = pango_attr_size_new(18000); attr->start_index = 0; attr->end_index = G_MAXINT; /* 18pt */
  pango_attr_list_insert(attr_list, attr);
  attr = pango_attr_foreground_new(0, 0, 65535); attr->start_index = 0; attr->end_index = G_MAXINT;
  pango_attr_list_insert(attr_list, attr);
  gtk_label_set_attributes(game->state_label, attr_list);
  pango_attr_list_unref(attr_list);
  
  init_colors(game);
  for (key = 0; key < NKEY; key++) {
    game->cur_un_idxs[key] = game->song->first_un_idxs[key];
    game->cur_un_idxs_long[key] = game->song->first_un_idxs_long[key];
  }
  game->speed = default_speed; game->pev_time_unit = default_pev_time_unit; /* 1ms */
  game->sched = new_event_scheduler(); game->recent_scores = g_new0(GameScore, recent_scores_nperiod);
  tempoit_init(game->song, &game->it);
  for (i = 0; i < game->song->ntrack; i++) {
    MidiTrack *track = &game->song->tracks[i];
    if (0 < track->nevent) {
      MidiEvent *ev = &track->events[0];
      double rt = tempoit_tt_to_rt(game->song, &game->it, ev->tt);
      sched_event(game->sched, rt, SchedEventMidi, i, 0);
    }
  }
  for (i = 0; i < game->song->num_note_array; i++) {
    NoteArray *arr = &game->song->note_arrays[i];
    if (0 < arr->num_note) {
      Note *note = &arr->notes[0];
      double rt = tempoit_tt_to_rt(game->song, &game->it, note->tt_start);
      sched_event(game->sched, rt, SchedEventNoteOn, i, 0);
    }
  }
  if (game->mode == VosGameModePlayDemo && 0 < game->in_nev) {
    InputEvent *ev = &game->in_evs[0];
    sched_event(game->sched, ev->rt, SchedEventDemo, 0, 0);
  }
  sched_event(game->sched, 0.0, SchedEventRefreshScore, 0, 0);
  if (game->mode == VosGameModeAuto) {
    for (i = 0; i < NKEY; i++) {
      unsigned idx = game->song->first_un_idxs[i];
      if (idx != IDX_NONE)
	sched_event(game->sched, game->song->user_notes[idx].rt_start, SchedEventAutoOn, i, idx);
    }
  }
  for (i = 0; i < NKEY; i++) {
    KeyState *ks = &game->keys[i];
    ks->key_down = FALSE; ks->un_idx = IDX_NONE;
    ks->cmp_press.count0 = 0; ks->cmp_press.score0 = 0.0;
    ks->cmp_press.nent = 0; ks->cmp_press.ents = NULL;
    ks->cmp_release.count0 = 0; ks->cmp_release.score0 = 0.0;
    ks->cmp_release.nent = 0; ks->cmp_release.ents = NULL;
    ks->long_score = 0.0; ks->total_long_score = 0.0;
    ks->markers = g_queue_new();
  }
  game->input_events_arr = g_array_new(FALSE, FALSE, sizeof(InputEvent));

  /* Start the game */
  g_get_current_time(&game->start_time); g_time_val_add(&game->start_time, 1000000); /* Start 1sec later */
  game->is_paused = FALSE; game->start_rt = 0.0;
  process_events(game); /* Add the event timeout */
  game->conn_key_press = g_signal_connect(G_OBJECT(game->main_area), "key-press-event",
					  G_CALLBACK(on_main_area_key), game);
  game->conn_key_release = g_signal_connect(G_OBJECT(game->main_area), "key-release-event",
					    G_CALLBACK(on_main_area_key), game);
  game->conn_expose = g_signal_connect(G_OBJECT(game->main_area), "expose-event",
				       G_CALLBACK(on_main_area_expose), game);
  game->refresh_timeout_id = gtk_timeout_add(5, on_refresh_timeout, game); /* 200fps */
  game->conn_delete = g_signal_connect(G_OBJECT(main_win), "delete-event",
				       G_CALLBACK(on_main_window_delete), game);
  update_game_state(game); update_game_status(game); update_game_score(game, get_rt(game));
  if (! fast_mode) {
    gtk_widget_show(main_win);
    on_refresh_timeout(game);
    gtk_main();
  }

  refresh_scores(game, get_rt(game)); get_game_score(game, &sc);
  g_message("Final score: %0.2f / %0.2f (%0.2f%%)", sc.score, sc.total_score,
	    (sc.total_score == 0.0) ? 100.0 : (sc.score / sc.total_score * 100.0));
  if (game->event_timeout_id) { gtk_timeout_remove(game->event_timeout_id); game->event_timeout_id = 0; }
  if (game->refresh_timeout_id) {
    gtk_timeout_remove(game->refresh_timeout_id); game->refresh_timeout_id = 0;
  }
  g_signal_handler_disconnect(G_OBJECT(main_win), game->conn_delete);
  g_signal_handler_disconnect(G_OBJECT(game->main_area), game->conn_expose);
  g_signal_handler_disconnect(G_OBJECT(game->main_area), game->conn_key_release);
  g_signal_handler_disconnect(G_OBJECT(game->main_area), game->conn_key_press);

  /* Save and free the event list */
  save_input_events(game, pev_out_fname);
  
  for (i = 0; i < NKEY; i++) {
    KeyState *ks = &game->keys[i];
    GQueue *markers = ks->markers;
    Marker *m;

    while ((m = (Marker *) g_queue_pop_head(markers)) != NULL) g_free(m);
    g_queue_free(markers);
    g_free(ks->cmp_press.ents); g_free(ks->cmp_release.ents);
  }
  g_free(game->recent_scores); free_event_scheduler(game->sched);
  g_free(game->colors);
  g_object_unref(G_OBJECT(gx));
  free_vos_song(game->song);
  result = snd_seq_close(game->seq); g_assert(result == 0);
  if (game->in_evs) g_free(game->in_evs);
  g_free(game);
  poptFreeContext(popt_ctx);
    
  return 0;
}
